@@ -1,27 +1,29 @@
+use crate::error::{self, ExpectedKind, Received, is_eof_io_error, io_syntax_error};
+
 use num::bigint::BigInt;
 use num::traits::cast::{FromPrimitive, ToPrimitive};
+
 use std::borrow::Cow;
 use std::convert::TryFrom;
 use std::convert::TryInto;
+use std::io;
 use std::marker::PhantomData;
 
 use super::constants::Tag;
 use super::super::{
-    FALSE,
-    IOResult,
-    IOValue,
+    CompoundClass,
+    DomainDecode,
+    Embeddable,
     Map,
     NestedValue,
     Record,
     Set,
-    TRUE,
     Value,
 
+    boundary as B,
     reader::{
+        Token,
         BinarySource,
-        BytesBinarySource,
-        ConfiguredReader,
-        IOBinarySource,
         Reader,
         ReaderResult,
     },
@@ -29,24 +31,33 @@ use super::super::{
     signed_integer::SignedInteger,
 };
 
-use crate::error::{self, ExpectedKind, Received, is_eof_io_error, io_syntax_error};
-
-pub struct PackedReader<'de, S: BinarySource<'de>> {
-    pub source: S,
-    phantom: PhantomData<&'de ()>,
+pub struct PackedReader<'de, 'src, D: Embeddable, N: NestedValue<D>, Dec: DomainDecode<D>, S: BinarySource<'de>> {
+    pub source: &'src mut S,
+    pub decode_embedded: Dec,
+    phantom: PhantomData<&'de (D, N)>,
 }
 
-impl<'de, S: BinarySource<'de>> BinarySource<'de> for PackedReader<'de, S> {
-    fn skip(&mut self) -> IOResult<()> {
+impl<'de, 'src, D: Embeddable, N: NestedValue<D>, Dec: DomainDecode<D>, S: BinarySource<'de>>
+    BinarySource<'de>
+    for PackedReader<'de, 'src, D, N, Dec, S>
+{
+    type Mark = S::Mark;
+    fn mark(&mut self) -> io::Result<Self::Mark> {
+        self.source.mark()
+    }
+    fn restore(&mut self, mark: &Self::Mark) -> io::Result<()> {
+        self.source.restore(mark)
+    }
+    fn skip(&mut self) -> io::Result<()> {
         self.source.skip()
     }
-    fn peek(&mut self) -> IOResult<u8> {
+    fn peek(&mut self) -> io::Result<u8> {
         self.source.peek()
     }
-    fn readbytes(&mut self, count: usize) -> IOResult<Cow<'de, [u8]>> {
+    fn readbytes(&mut self, count: usize) -> io::Result<Cow<'de, [u8]>> {
         self.source.readbytes(count)
     }
-    fn readbytes_into(&mut self, bs: &mut [u8]) -> IOResult<()> {
+    fn readbytes_into(&mut self, bs: &mut [u8]) -> io::Result<()> {
         self.source.readbytes_into(bs)
     }
 }
@@ -55,24 +66,12 @@ fn out_of_range<I: Into<BigInt>>(i: I) -> error::Error {
     error::Error::NumberOutOfRange(i.into())
 }
 
-impl<'de> PackedReader<'de, BytesBinarySource<'de>> {
-    pub fn decode_bytes(bytes: &'de [u8]) -> Self {
-        PackedReader::new(BytesBinarySource::new(bytes))
-    }
-}
-
-impl<'de, 'a, IOR: std::io::Read> PackedReader<'de, IOBinarySource<'a, IOR>> {
-    pub fn decode_read(read: &'a mut IOR) -> Self {
-        PackedReader::new(IOBinarySource::new(read))
-    }
-}
-
-impl<'de, S: BinarySource<'de>> PackedReader<'de, S> {
-    pub fn new(source: S) -> Self {
-        PackedReader { source, phantom: PhantomData }
+impl<'de, 'src, D: Embeddable, N: NestedValue<D>, Dec: DomainDecode<D>, S: BinarySource<'de>> PackedReader<'de, 'src, D, N, Dec, S> {
+    pub fn new(source: &'src mut S, decode_embedded: Dec) -> Self {
+        PackedReader { source, decode_embedded, phantom: PhantomData }
     }
 
-    fn read(&mut self) -> IOResult<u8> {
+    fn read(&mut self) -> io::Result<u8> {
         let v = self.peek()?;
         self.skip()?;
         Ok(v)
@@ -80,12 +79,12 @@ impl<'de, S: BinarySource<'de>> PackedReader<'de, S> {
 
     fn expected(&mut self, k: ExpectedKind) -> error::Error {
         match self.demand_next(true) {
-            Ok(v) => error::Error::Expected(k, Received::ReceivedOtherValue(v)),
+            Ok(v) => error::Error::Expected(k, Received::ReceivedOtherValue(format!("{:?}", v))),
             Err(e) => e.into()
         }
     }
 
-    fn varint(&mut self) -> IOResult<usize> {
+    fn varint(&mut self) -> io::Result<usize> {
         let mut shift = 0;
         let mut acc: usize = 0;
         loop {
@@ -96,7 +95,7 @@ impl<'de, S: BinarySource<'de>> PackedReader<'de, S> {
         }
     }
 
-    fn peekend(&mut self) -> IOResult<bool> {
+    fn peekend(&mut self) -> io::Result<bool> {
         if self.peek()? == Tag::End.into() {
             self.skip()?;
             Ok(true)
@@ -139,7 +138,7 @@ impl<'de, S: BinarySource<'de>> PackedReader<'de, S> {
         }
     }
 
-    fn read_signed_integer(&mut self, count: usize) -> IOResult<SignedInteger> {
+    fn read_signed_integer(&mut self, count: usize) -> io::Result<SignedInteger> {
         if count == 0 {
             return Ok(SignedInteger::from(0_i128));
         }
@@ -232,18 +231,46 @@ impl<'de, S: BinarySource<'de>> PackedReader<'de, S> {
             _ => Err(self.expected(ExpectedKind::SignedInteger))
         }
     }
+
+    fn gather_annotations(&mut self) -> io::Result<Vec<N>> {
+        let mut annotations = vec![self.demand_next(true)?];
+        while Tag::try_from(self.peek()?)? == Tag::Annotation {
+            self.skip()?;
+            annotations.push(self.demand_next(true)?);
+        }
+        Ok(annotations)
+    }
+
+    fn skip_annotations(&mut self) -> io::Result<()> {
+        self.skip_value()?;
+        while Tag::try_from(self.peek()?)? == Tag::Annotation {
+            self.skip()?;
+            self.skip_value()?;
+        }
+        Ok(())
+    }
+
+    fn next_upto_end(&mut self, read_annotations: bool) -> io::Result<Option<N>> {
+        match self.peekend()? {
+            true => Ok(None),
+            false => Ok(Some(self.demand_next(read_annotations)?)),
+        }
+    }
 }
 
-impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
-    fn next(&mut self, read_annotations: bool) -> IOResult<Option<IOValue>> {
+impl<'de, 'src, D: Embeddable, N: NestedValue<D>, Dec: DomainDecode<D>, S: BinarySource<'de>>
+    Reader<'de, D, N>
+    for PackedReader<'de, 'src, D, N, Dec, S>
+{
+    fn next(&mut self, read_annotations: bool) -> io::Result<Option<N>> {
         match self.peek() {
             Err(e) if is_eof_io_error(&e) => return Ok(None),
             Err(e) => return Err(e),
             Ok(_) => (),
         }
         Ok(Some(match Tag::try_from(self.read()?)? {
-            Tag::False => FALSE.clone(),
-            Tag::True => TRUE.clone(),
+            Tag::False => N::new(false),
+            Tag::True => N::new(true),
             Tag::Float => {
                 let mut bs = [0; 4];
                 self.readbytes_into(&mut bs)?;
@@ -256,26 +283,17 @@ impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
             }
             Tag::Annotation => {
                 if read_annotations {
-                    let mut annotations = vec![self.demand_next(read_annotations)?];
-                    while Tag::try_from(self.peek()?)? == Tag::Annotation {
-                        self.skip()?;
-                        annotations.push(self.demand_next(read_annotations)?);
-                    }
+                    let mut annotations = self.gather_annotations()?;
                     let (existing_annotations, v) = self.demand_next(read_annotations)?.pieces();
                     annotations.extend_from_slice(existing_annotations.slice());
-                    IOValue::wrap(Annotations::new(Some(annotations)), v)
+                    N::wrap(Annotations::new(Some(annotations)), v)
                 } else {
-                    self.skip_value()?;
-                    while Tag::try_from(self.peek()?)? == Tag::Annotation {
-                        self.skip()?;
-                        self.skip_value()?;
-                    }
+                    self.skip_annotations()?;
                     self.demand_next(read_annotations)?
                 }
             }
             Tag::Embedded => {
-                let v = self.demand_next(read_annotations)?;
-                Value::Domain(v).wrap()
+                Value::Embedded(self.decode_embedded.decode_embedded(self.source, read_annotations)?).wrap()
             }
             Tag::SmallInteger(v) => {
                 // TODO: prebuild these in value.rs
@@ -303,34 +321,28 @@ impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
                 Value::Symbol(decodestr(self.readbytes(count)?)?.into_owned()).wrap()
             }
             Tag::Record => {
-                let iter = DelimitedStream { reader: self.configured(read_annotations) };
-                let vs = iter.collect::<IOResult<Vec<IOValue>>>()?;
+                let mut vs = Vec::new();
+                while let Some(v) = self.next_upto_end(read_annotations)? { vs.push(v); }
                 if vs.is_empty() {
                     return Err(io_syntax_error("Too few elements in encoded record"))
                 }
                 Value::Record(Record(vs)).wrap()
             }
             Tag::Sequence => {
-                let iter = DelimitedStream { reader: self.configured(read_annotations) };
-                let vs = iter.collect::<IOResult<Vec<IOValue>>>()?;
+                let mut vs = Vec::new();
+                while let Some(v) = self.next_upto_end(read_annotations)? { vs.push(v); }
                 Value::Sequence(vs).wrap()
             }
             Tag::Set => {
-                let iter = DelimitedStream { reader: self.configured(read_annotations) };
                 let mut s = Set::new();
-                for res in iter { s.insert(res?); }
+                while let Some(v) = self.next_upto_end(read_annotations)? { s.insert(v); }
                 Value::Set(s).wrap()
             }
             Tag::Dictionary => {
-                let mut iter = DelimitedStream { reader: self.configured(read_annotations) };
                 let mut d = Map::new();
-                while let Some(kres) = iter.next() {
-                    let k = kres?;
-                    match iter.next() {
-                        Some(vres) => {
-                            let v = vres?;
-                            d.insert(k, v);
-                        }
+                while let Some(k) = self.next_upto_end(read_annotations)? {
+                    match self.next_upto_end(read_annotations)? {
+                        Some(v) => { d.insert(k, v); }
                         None => return Err(io_syntax_error("Missing dictionary value")),
                     }
                 }
@@ -342,18 +354,25 @@ impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
         }))
     }
 
-    fn open_record(&mut self, arity: Option<usize>) -> ReaderResult<()> {
+    fn open_record(&mut self, arity: Option<usize>) -> ReaderResult<B::Type> {
         self.next_compound(Tag::Record, ExpectedKind::Record(arity))?;
-        self.ensure_more_expected()
+        let mut b = B::Type::default();
+        self.ensure_more_expected(&mut b, &B::Item::RecordLabel)?;
+        Ok(b)
     }
 
-    fn open_sequence_or_set(&mut self) -> ReaderResult<()> {
+    fn open_sequence_or_set(&mut self) -> ReaderResult<B::Item> {
         match self.peek_next_nonannotation_tag()? {
-            Tag::Sequence | Tag::Set => {
+            Tag::Sequence => {
                 self.skip()?;
-                Ok(())
+                Ok(B::Item::SequenceValue)
             }
-            _ => Err(self.expected(ExpectedKind::SequenceOrSet)),
+            Tag::Set => {
+                self.skip()?;
+                Ok(B::Item::SetValue)
+            }
+            _ =>
+                Err(self.expected(ExpectedKind::SequenceOrSet)),
         }
     }
 
@@ -369,7 +388,12 @@ impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
         self.next_compound(Tag::Dictionary, ExpectedKind::Dictionary)
     }
 
-    fn close_compound(&mut self) -> ReaderResult<bool> {
+    #[inline]
+    fn boundary(&mut self, _b: &B::Type) -> ReaderResult<()> {
+        Ok(())
+    }
+
+    fn close_compound(&mut self, _b: &mut B::Type, _i: &B::Item) -> ReaderResult<bool> {
         Ok(self.peekend()?)
     }
 
@@ -381,6 +405,64 @@ impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
         Ok(())
     }
 
+    type Mark = S::Mark;
+
+    fn mark(&mut self) -> io::Result<Self::Mark> {
+        self.source.mark()
+    }
+
+    fn restore(&mut self, mark: &Self::Mark) -> io::Result<()> {
+        self.source.restore(mark)
+    }
+
+    fn next_token(&mut self, read_embedded_annotations: bool) -> io::Result<Token<D, N>> {
+        loop {
+            return Ok(match Tag::try_from(self.peek()?)? {
+                Tag::Embedded => {
+                    self.skip()?;
+                    Token::Embedded(self.decode_embedded.decode_embedded(
+                        self.source,
+                        read_embedded_annotations)?)
+                }
+                Tag::False |
+                Tag::True |
+                Tag::Float |
+                Tag::Double |
+                Tag::SmallInteger(_) |
+                Tag::MediumInteger(_) |
+                Tag::SignedInteger |
+                Tag::String |
+                Tag::ByteString |
+                Tag::Symbol =>
+                    Token::Atom(self.demand_next(false)?),
+
+                Tag::Record => { self.skip()?; Token::Compound(CompoundClass::Record) }
+                Tag::Sequence => { self.skip()?; Token::Compound(CompoundClass::Sequence) }
+                Tag::Set => { self.skip()?; Token::Compound(CompoundClass::Set) }
+                Tag::Dictionary => { self.skip()?; Token::Compound(CompoundClass::Dictionary) }
+
+                Tag::End => { self.skip()?; Token::End }
+
+                Tag::Annotation => {
+                    self.skip()?;
+                    self.skip_annotations()?;
+                    continue
+                }
+            })
+        }
+    }
+
+    fn next_annotations_and_token(&mut self) -> io::Result<(Vec<N>, Token<D, N>)> {
+        match Tag::try_from(self.peek()?)? {
+            Tag::Annotation => {
+                self.skip()?;
+                let annotations = self.gather_annotations()?;
+                Ok((annotations, self.next_token(true)?))
+            }
+            _ => Ok((Vec::new(), self.next_token(true)?)),
+        }
+    }
+
     fn next_boolean(&mut self) -> ReaderResult<bool> {
         match self.peek_next_nonannotation_tag()? {
             Tag::False => { self.skip()?; Ok(false) }
@@ -389,6 +471,26 @@ impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
         }
     }
 
+    fn next_signedinteger(&mut self) -> ReaderResult<SignedInteger> {
+        let tag = self.peek_next_nonannotation_tag()?;
+        match tag {
+            Tag::SmallInteger(v) => {
+                self.skip()?;
+                Ok(SignedInteger::from(v as i32))
+            }
+            Tag::MediumInteger(count) => {
+                self.skip()?;
+                Ok(self.read_signed_integer(count.into())?)
+            }
+            Tag::SignedInteger => {
+                self.skip()?;
+                let count = self.varint()?;
+                Ok(self.read_signed_integer(count)?)
+            }
+            _ => Err(self.expected(ExpectedKind::SignedInteger))
+        }
+    }
+
     fn next_i8(&mut self)  -> ReaderResult<i8>    { self.next_signed(|n| n.to_i8()) }
     fn next_i16(&mut self) -> ReaderResult<i16>   { self.next_signed(|n| n.to_i16()) }
     fn next_i32(&mut self) -> ReaderResult<i32>   { self.next_signed(|n| n.to_i32()) }
@@ -450,23 +552,7 @@ impl<'de, S: BinarySource<'de>> Reader<'de> for PackedReader<'de, S> {
     }
 }
 
-struct DelimitedStream<'a, 'de, S: BinarySource<'de>> {
-    reader: ConfiguredReader<'de, &'a mut PackedReader<'de, S>>,
-}
-
-impl<'a, 'de, S: BinarySource<'de>> Iterator for DelimitedStream<'a, 'de, S>
-{
-    type Item = IOResult<IOValue>;
-    fn next(&mut self) -> Option<Self::Item> {
-        match self.reader.reader.peekend() {
-            Err(e) => Some(Err(e)),
-            Ok(true) => None,
-            Ok(false) => Some(self.reader.reader.demand_next(self.reader.read_annotations)),
-        }
-    }
-}
-
-fn decodestr(cow: Cow<'_, [u8]>) -> IOResult<Cow<'_, str>> {
+fn decodestr(cow: Cow<'_, [u8]>) -> io::Result<Cow<'_, str>> {
     match cow {
         Cow::Borrowed(bs) =>
             Ok(Cow::Borrowed(std::str::from_utf8(bs).map_err(|_| io_syntax_error("Invalid UTF-8"))?)),
