@@ -17,26 +17,24 @@ use std::vec::Vec;
 pub use std::collections::BTreeSet as Set;
 pub use std::collections::BTreeMap as Map;
 
-use super::IOResult;
 use super::signed_integer::SignedInteger;
 use crate::error::{Error, ExpectedKind, Received};
 
-pub trait Domain: Sized + Debug + Clone + Eq + Hash + Ord {
-    fn from_preserves(v: IOValue) -> IOResult<Self>;
-    fn as_preserves(&self) -> IOValue;
-}
+pub trait Domain: Sized + Debug + Eq + Hash + Ord {}
+pub trait Embeddable: Domain + Clone {}
+impl<T> Embeddable for T where T: Domain + Clone {}
 
-impl<D: Domain> Domain for Arc<D> {
-    fn from_preserves(v: IOValue) -> IOResult<Self> {
-        Ok(Arc::new(D::from_preserves(v)?))
+impl<D: Domain> Domain for Arc<D> {}
+
+pub trait NestedValue<D: Embeddable>: Sized + Debug + Clone + Eq + Hash + Ord {
+    fn new<V>(v: V) -> Self where Value<Self, D>: From<V> {
+        Value::from(v).wrap()
     }
 
-    fn as_preserves(&self) -> IOValue {
-        D::as_preserves(self)
+    fn domain<E>(e: E) -> Self where D: From<E> {
+        Value::Embedded(e.into()).wrap()
     }
-}
 
-pub trait NestedValue<D: Domain>: Sized + Debug + Clone + Eq + Hash + Ord {
     fn wrap(anns: Annotations<Self, D>, v: Value<Self, D>) -> Self;
 
     fn annotations(&self) -> &Annotations<Self, D>;
@@ -44,6 +42,10 @@ pub trait NestedValue<D: Domain>: Sized + Debug + Clone + Eq + Hash + Ord {
     fn pieces(self) -> (Annotations<Self, D>, Value<Self, D>);
     fn value_owned(self) -> Value<Self, D>;
 
+    fn value_class(&self) -> ValueClass {
+        self.value().value_class()
+    }
+
     fn debug_fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         for ann in self.annotations().slice() {
             write!(f, "@{:?} ", ann)?;
@@ -51,25 +53,28 @@ pub trait NestedValue<D: Domain>: Sized + Debug + Clone + Eq + Hash + Ord {
         self.value().fmt(f)
     }
 
-    fn copy_via<M: NestedValue<E>, E: Domain, F>(&self, f: &F) -> IOResult<M>
+    fn copy_via<M: NestedValue<E>, E: Embeddable, F, Err>(&self, f: &mut F) -> Result<M, Err>
     where
-        F: Fn(&D) -> IOResult<Value<M, E>>
+        F: FnMut(&D) -> Result<Value<M, E>, Err>
     {
         Ok(M::wrap(self.annotations().copy_via(f)?, self.value().copy_via(f)?))
     }
 
-    fn to_io_value(&self) -> IOValue {
-        self.copy_via(&|d| Ok(d.as_preserves().value().clone())).unwrap()
-    }
-
-    fn from_io_value(v: &IOValue) -> IOResult<Self> {
-        v.copy_via(&|d| Ok(Value::Domain(D::from_preserves(d.clone())?)))
+    fn foreach_embedded<F, Err>(&self, f: &mut F) -> Result<(), Err>
+    where
+        F: FnMut(&D) -> Result<(), Err>
+    {
+        match &self.annotations().0 {
+            None => (),
+            Some(vs) => for v in vs { v.foreach_embedded(f)? },
+        }
+        self.value().foreach_embedded(f)
     }
 }
 
 /// The `Value`s from the specification.
 #[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
-pub enum Value<N, D> where N: NestedValue<D>, D: Domain {
+pub enum Value<N, D> where N: NestedValue<D>, D: Embeddable {
     Boolean(bool),
     Float(Float),
     Double(Double),
@@ -81,15 +86,44 @@ pub enum Value<N, D> where N: NestedValue<D>, D: Domain {
     Sequence(Vec<N>),
     Set(Set<N>),
     Dictionary(Map<N, N>),
-    Domain(D),
+    Embedded(D),
+}
+
+/// The kinds of `Value` from the specification.
+#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
+pub enum ValueClass {
+    Atomic(AtomClass),
+    Compound(CompoundClass),
+    Embedded,
+}
+
+/// The kinds of `Atom` from the specification.
+#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
+pub enum AtomClass {
+    Boolean,
+    Float,
+    Double,
+    SignedInteger,
+    String,
+    ByteString,
+    Symbol,
+}
+
+/// The kinds of `Compound` from the specification.
+#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
+pub enum CompoundClass {
+    Record,
+    Sequence,
+    Set,
+    Dictionary,
 }
 
 /// Single-precision IEEE 754 Value
-#[derive(Clone, Debug)]
+#[derive(Clone, Copy, Debug)]
 pub struct Float(pub f32);
 
 /// Double-precision IEEE 754 Value
-#[derive(Clone, Debug)]
+#[derive(Clone, Copy, Debug)]
 pub struct Double(pub f64);
 
 /// A Record `Value` -- INVARIANT: length always non-zero
@@ -125,7 +159,7 @@ impl<N> Record<N> {
         &mut self.0
     }
 
-    pub fn finish<D: Domain>(self) -> Value<N, D> where N: NestedValue<D> {
+    pub fn finish<D: Embeddable>(self) -> Value<N, D> where N: NestedValue<D> {
         Value::Record(self)
     }
 }
@@ -214,38 +248,41 @@ impl PartialOrd for Double {
 
 impl Eq for Double {}
 
-impl<N: NestedValue<D>, D: Domain> From<bool> for Value<N, D> { fn from(v: bool) -> Self { Value::Boolean(v) } }
+impl<N: NestedValue<D>, D: Embeddable> From<bool> for Value<N, D> { fn from(v: bool) -> Self { Value::Boolean(v) } }
 
-impl<N: NestedValue<D>, D: Domain> From<f32> for Value<N, D> { fn from(v: f32) -> Self { Value::Float(Float::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<&Float> for Value<N, D> { fn from(v: &Float) -> Self { Value::Float(v.clone()) } }
-impl<N: NestedValue<D>, D: Domain> From<f64> for Value<N, D> { fn from(v: f64) -> Self { Value::Double(Double::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<&Double> for Value<N, D> { fn from(v: &Double) -> Self { Value::Double(v.clone()) } }
+impl<N: NestedValue<D>, D: Embeddable> From<f32> for Value<N, D> { fn from(v: f32) -> Self { Value::Float(Float::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<&Float> for Value<N, D> { fn from(v: &Float) -> Self { Value::Float(v.clone()) } }
+impl<N: NestedValue<D>, D: Embeddable> From<f64> for Value<N, D> { fn from(v: f64) -> Self { Value::Double(Double::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<&Double> for Value<N, D> { fn from(v: &Double) -> Self { Value::Double(v.clone()) } }
 
-impl<N: NestedValue<D>, D: Domain> From<u8> for Value<N, D> { fn from(v: u8) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<i8> for Value<N, D> { fn from(v: i8) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<u16> for Value<N, D> { fn from(v: u16) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<i16> for Value<N, D> { fn from(v: i16) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<u32> for Value<N, D> { fn from(v: u32) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<i32> for Value<N, D> { fn from(v: i32) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<u64> for Value<N, D> { fn from(v: u64) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<i64> for Value<N, D> { fn from(v: i64) -> Self { Value::from(i128::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<u128> for Value<N, D> { fn from(v: u128) -> Self { Value::SignedInteger(SignedInteger::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<i128> for Value<N, D> { fn from(v: i128) -> Self { Value::SignedInteger(SignedInteger::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<&BigInt> for Value<N, D> { fn from(v: &BigInt) -> Self { Value::SignedInteger(SignedInteger::from(Cow::Borrowed(v))) } }
-impl<N: NestedValue<D>, D: Domain> From<&SignedInteger> for Value<N, D> { fn from(v: &SignedInteger) -> Self { Value::SignedInteger(v.clone()) } }
+impl<N: NestedValue<D>, D: Embeddable> From<u8> for Value<N, D> { fn from(v: u8) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<i8> for Value<N, D> { fn from(v: i8) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<u16> for Value<N, D> { fn from(v: u16) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<i16> for Value<N, D> { fn from(v: i16) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<u32> for Value<N, D> { fn from(v: u32) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<i32> for Value<N, D> { fn from(v: i32) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<u64> for Value<N, D> { fn from(v: u64) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<i64> for Value<N, D> { fn from(v: i64) -> Self { Value::from(i128::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<usize> for Value<N, D> { fn from(v: usize) -> Self { Value::from(v as u128) } }
+impl<N: NestedValue<D>, D: Embeddable> From<isize> for Value<N, D> { fn from(v: isize) -> Self { Value::from(v as i128) } }
+impl<N: NestedValue<D>, D: Embeddable> From<u128> for Value<N, D> { fn from(v: u128) -> Self { Value::SignedInteger(SignedInteger::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<i128> for Value<N, D> { fn from(v: i128) -> Self { Value::SignedInteger(SignedInteger::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<&BigInt> for Value<N, D> { fn from(v: &BigInt) -> Self { Value::SignedInteger(SignedInteger::from(Cow::Borrowed(v))) } }
+impl<N: NestedValue<D>, D: Embeddable> From<BigInt> for Value<N, D> { fn from(v: BigInt) -> Self { Value::SignedInteger(SignedInteger::from(Cow::Owned(v))) } }
+impl<N: NestedValue<D>, D: Embeddable> From<&SignedInteger> for Value<N, D> { fn from(v: &SignedInteger) -> Self { Value::SignedInteger(v.clone()) } }
 
-impl<N: NestedValue<D>, D: Domain> From<&str> for Value<N, D> { fn from(v: &str) -> Self { Value::String(String::from(v)) } }
-impl<N: NestedValue<D>, D: Domain> From<String> for Value<N, D> { fn from(v: String) -> Self { Value::String(v) } }
-impl<N: NestedValue<D>, D: Domain> From<&String> for Value<N, D> { fn from(v: &String) -> Self { Value::String(v.to_owned()) } }
+impl<N: NestedValue<D>, D: Embeddable> From<&str> for Value<N, D> { fn from(v: &str) -> Self { Value::String(String::from(v)) } }
+impl<N: NestedValue<D>, D: Embeddable> From<String> for Value<N, D> { fn from(v: String) -> Self { Value::String(v) } }
+impl<N: NestedValue<D>, D: Embeddable> From<&String> for Value<N, D> { fn from(v: &String) -> Self { Value::String(v.to_owned()) } }
 
-impl<N: NestedValue<D>, D: Domain> From<&[u8]> for Value<N, D> { fn from(v: &[u8]) -> Self { Value::ByteString(Vec::from(v)) } }
-// impl<N: NestedValue<D>, D: Domain> From<Vec<u8>> for Value<N, D> { fn from(v: Vec<u8>) -> Self { Value::ByteString(v) } }
+impl<N: NestedValue<D>, D: Embeddable> From<&[u8]> for Value<N, D> { fn from(v: &[u8]) -> Self { Value::ByteString(Vec::from(v)) } }
+// impl<N: NestedValue<D>, D: Embeddable> From<Vec<u8>> for Value<N, D> { fn from(v: Vec<u8>) -> Self { Value::ByteString(v) } }
 
-impl<N: NestedValue<D>, D: Domain> From<Vec<N>> for Value<N, D> { fn from(v: Vec<N>) -> Self { Value::Sequence(v) } }
-impl<N: NestedValue<D>, D: Domain> From<Set<N>> for Value<N, D> { fn from(v: Set<N>) -> Self { Value::Set(v) } }
-impl<N: NestedValue<D>, D: Domain> From<Map<N, N>> for Value<N, D> { fn from(v: Map<N, N>) -> Self { Value::Dictionary(v) } }
+impl<N: NestedValue<D>, D: Embeddable> From<Vec<N>> for Value<N, D> { fn from(v: Vec<N>) -> Self { Value::Sequence(v) } }
+impl<N: NestedValue<D>, D: Embeddable> From<Set<N>> for Value<N, D> { fn from(v: Set<N>) -> Self { Value::Set(v) } }
+impl<N: NestedValue<D>, D: Embeddable> From<Map<N, N>> for Value<N, D> { fn from(v: Map<N, N>) -> Self { Value::Dictionary(v) } }
 
-impl<N: NestedValue<D>, D: Domain> Debug for Value<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> Debug for Value<N, D> {
     // Not *quite* a formatter for the Preserves text syntax, since it
     // doesn't escape strings/symbols properly.
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
@@ -284,20 +321,55 @@ impl<N: NestedValue<D>, D: Domain> Debug for Value<N, D> {
                 f.debug_set().entries(v.iter()).finish()
             }
             Value::Dictionary(v) => f.debug_map().entries(v.iter()).finish(),
-            Value::Domain(d) => write!(f, "#!{:?}", d),
+            Value::Embedded(d) => write!(f, "#!{:?}", d),
         }
     }
 }
 
 //---------------------------------------------------------------------------
 
-impl<N: NestedValue<D>, D: Domain> Value<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> Value<N, D> {
     pub fn wrap(self) -> N {
         N::wrap(Annotations::empty(), self)
     }
 
+    fn value_class(&self) -> ValueClass {
+        match self {
+            Value::Boolean(_) => ValueClass::Atomic(AtomClass::Boolean),
+            Value::Float(_) => ValueClass::Atomic(AtomClass::Float),
+            Value::Double(_) => ValueClass::Atomic(AtomClass::Double),
+            Value::SignedInteger(_) => ValueClass::Atomic(AtomClass::SignedInteger),
+            Value::String(_) => ValueClass::Atomic(AtomClass::String),
+            Value::ByteString(_) => ValueClass::Atomic(AtomClass::ByteString),
+            Value::Symbol(_) => ValueClass::Atomic(AtomClass::Symbol),
+            Value::Record(_) => ValueClass::Compound(CompoundClass::Record),
+            Value::Sequence(_) => ValueClass::Compound(CompoundClass::Sequence),
+            Value::Set(_) => ValueClass::Compound(CompoundClass::Set),
+            Value::Dictionary(_) => ValueClass::Compound(CompoundClass::Dictionary),
+            Value::Embedded(_) => ValueClass::Embedded,
+        }
+    }
+
+    pub fn children(&self) -> Vec<N> {
+        match self {
+            Value::Boolean(_) |
+            Value::Float(_) |
+            Value::Double(_) |
+            Value::SignedInteger(_) |
+            Value::String(_) |
+            Value::ByteString(_) |
+            Value::Symbol(_) |
+            Value::Embedded(_) => vec![],
+
+            Value::Record(r) => r.fields().to_vec(),
+            Value::Sequence(vs) => vs.clone(),
+            Value::Set(s) => s.iter().cloned().collect(),
+            Value::Dictionary(d) => d.values().cloned().collect(),
+        }
+    }
+
     fn expected(&self, k: ExpectedKind) -> Error {
-        Error::Expected(k, Received::ReceivedOtherValue(self.clone().wrap().to_io_value()))
+        Error::Expected(k, Received::ReceivedOtherValue(format!("{:?}", self.clone().wrap())))
     }
 
     pub fn is_boolean(&self) -> bool {
@@ -462,6 +534,8 @@ impl<N: NestedValue<D>, D: Domain> Value<N, D> {
     pub fn as_i64(&self) -> Option<i64> { self.as_i().and_then(|i| i.to_i64()) }
     pub fn as_u128(&self) -> Option<u128> { self.as_u().and_then(|i| i.to_u128()) }
     pub fn as_i128(&self) -> Option<i128> { self.as_i().and_then(|i| i.to_i128()) }
+    pub fn as_usize(&self) -> Option<usize> { self.as_u().and_then(|i| i.to_usize()) }
+    pub fn as_isize(&self) -> Option<isize> { self.as_i().and_then(|i| i.to_isize()) }
 
     pub fn to_i8(&self) -> Result<i8, Error> {
         match self.as_i() {
@@ -533,6 +607,20 @@ impl<N: NestedValue<D>, D: Domain> Value<N, D> {
         }
     }
 
+    pub fn to_isize(&self) -> Result<isize, Error> {
+        match self.as_i() {
+            Some(i) => i.to_isize().ok_or_else(|| Error::NumberOutOfRange(BigInt::from(i))),
+            None => Err(self.expected(ExpectedKind::SignedInteger)),
+        }
+    }
+
+    pub fn to_usize(&self) -> Result<usize, Error> {
+        match self.as_u() {
+            Some(i) => i.to_usize().ok_or_else(|| Error::NumberOutOfRange(BigInt::from(i))),
+            None => Err(self.expected(ExpectedKind::SignedInteger)),
+        }
+    }
+
     pub fn to_char(&self) -> Result<char, Error> {
         let fs = self.to_simple_record("UnicodeScalar", Some(1))?;
         let c = fs[0].value().to_u32()?;
@@ -774,7 +862,7 @@ impl<N: NestedValue<D>, D: Domain> Value<N, D> {
     }
 
     pub fn as_embedded(&self) -> Option<&D> {
-        if let Value::Domain(d) = self {
+        if let Value::Embedded(d) = self {
             Some(d)
         } else {
             None
@@ -785,9 +873,9 @@ impl<N: NestedValue<D>, D: Domain> Value<N, D> {
         self.as_embedded().ok_or_else(|| self.expected(ExpectedKind::Embedded))
     }
 
-    pub fn copy_via<M: NestedValue<E>, E: Domain, F>(&self, f: &F) -> IOResult<Value<M, E>>
+    pub fn copy_via<M: NestedValue<E>, E: Embeddable, F, Err>(&self, f: &mut F) -> Result<Value<M, E>, Err>
     where
-        F: Fn(&D) -> IOResult<Value<M, E>>
+        F: FnMut(&D) -> Result<Value<M, E>, Err>
     {
         Ok(match self {
             Value::Boolean(b) => Value::Boolean(*b),
@@ -801,18 +889,41 @@ impl<N: NestedValue<D>, D: Domain> Value<N, D> {
                 Value::Record(Record(r.fields_vec().iter().map(|a| a.copy_via(f))
                                      .collect::<Result<Vec<_>, _>>()?)),
             Value::Sequence(v) => Value::Sequence(v.iter().map(|a| a.copy_via(f))
-                                                      .collect::<IOResult<Vec<_>>>()?),
+                                                      .collect::<Result<Vec<_>, _>>()?),
             Value::Set(v) => Value::Set(v.iter().map(|a| a.copy_via(f))
-                                            .collect::<IOResult<Set<_>>>()?),
+                                            .collect::<Result<Set<_>, _>>()?),
             Value::Dictionary(v) =>
                 Value::Dictionary(v.iter().map(|(a,b)| Ok((a.copy_via(f)?, b.copy_via(f)?)))
-                                  .collect::<IOResult<Map<_, _>>>()?),
-            Value::Domain(d) => f(d)?,
+                                  .collect::<Result<Map<_, _>, _>>()?),
+            Value::Embedded(d) => f(d)?,
         })
     }
+
+    pub fn foreach_embedded<F, Err>(&self, f: &mut F) -> Result<(), Err>
+    where
+        F: FnMut(&D) -> Result<(), Err>
+    {
+        match self {
+            Value::Boolean(_) |
+            Value::Float(_) |
+            Value::Double(_) |
+            Value::SignedInteger(_) |
+            Value::String(_) |
+            Value::ByteString(_) |
+            Value::Symbol(_) => Ok(()),
+            Value::Record(r) => Ok(for v in r.fields_vec() { v.foreach_embedded(f)? }),
+            Value::Sequence(vs) => Ok(for v in vs { v.foreach_embedded(f)? }),
+            Value::Set(vs) => Ok(for v in vs { v.foreach_embedded(f)? }),
+            Value::Dictionary(d) => Ok(for (k, v) in d {
+                k.foreach_embedded(f)?;
+                v.foreach_embedded(f)?;
+            }),
+            Value::Embedded(d) => f(d),
+        }
+    }
 }
 
-impl<N: NestedValue<D>, D: Domain> Index<usize> for Value<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> Index<usize> for Value<N, D> {
     type Output = N;
 
     fn index(&self, i: usize) -> &Self::Output {
@@ -820,13 +931,13 @@ impl<N: NestedValue<D>, D: Domain> Index<usize> for Value<N, D> {
     }
 }
 
-impl<N: NestedValue<D>, D: Domain> IndexMut<usize> for Value<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> IndexMut<usize> for Value<N, D> {
     fn index_mut(&mut self, i: usize) -> &mut Self::Output {
         &mut self.as_sequence_mut().unwrap()[i]
     }
 }
 
-impl<N: NestedValue<D>, D: Domain> Index<&N> for Value<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> Index<&N> for Value<N, D> {
     type Output = N;
 
     fn index(&self, i: &N) -> &Self::Output {
@@ -849,25 +960,12 @@ impl<'de> serde::Deserialize<'de> for UnwrappedIOValue {
     }
 }
 
-pub fn serialize_nested_value<S, N, D: Domain>(v: &N, serializer: S) ->
-    Result<S::Ok, S::Error> where S: serde::Serializer, N: NestedValue<D> + serde::Serialize
-{
-    super::magic::output_value(serializer, v.to_io_value())
-}
-
-pub fn deserialize_nested_value<'de, D, N, Dom: Domain>(deserializer: D) ->
-    Result<N, D::Error> where D: serde::Deserializer<'de>, N: NestedValue<Dom> + serde::Deserialize<'de>
-{
-    N::from_io_value(&super::magic::input_value(deserializer)?)
-        .map_err(|e| serde::de::Error::custom(format!("{:?}", e)))
-}
-
 //---------------------------------------------------------------------------
 
 #[derive(Clone)]
-pub struct Annotations<N: NestedValue<D>, D: Domain>(Option<Vec<N>>, PhantomData<D>);
+pub struct Annotations<N: NestedValue<D>, D: Embeddable>(Option<Vec<N>>, PhantomData<D>);
 
-impl<N: NestedValue<D>, D: Domain> Annotations<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> Annotations<N, D> {
     pub fn empty() -> Self {
         Annotations(None,
                     PhantomData)
@@ -898,9 +996,9 @@ impl<N: NestedValue<D>, D: Domain> Annotations<N, D> {
         self
     }
 
-    pub fn copy_via<M: NestedValue<E>, E: Domain, F>(&self, f: &F) -> IOResult<Annotations<M, E>>
+    pub fn copy_via<M: NestedValue<E>, E: Embeddable, F, Err>(&self, f: &mut F) -> Result<Annotations<M, E>, Err>
     where
-        F: Fn(&D) -> IOResult<Value<M, E>>
+        F: FnMut(&D) -> Result<Value<M, E>, Err>
     {
         Ok(match &self.0 {
             None =>
@@ -915,35 +1013,35 @@ impl<N: NestedValue<D>, D: Domain> Annotations<N, D> {
 /// A possibly-annotated Value, with annotations (themselves
 /// possibly-annotated) in order of appearance.
 #[derive(Clone)]
-pub struct AnnotatedValue<N: NestedValue<D>, D: Domain>(pub Annotations<N, D>, pub Value<N, D>);
+pub struct AnnotatedValue<N: NestedValue<D>, D: Embeddable>(pub Annotations<N, D>, pub Value<N, D>);
 
-impl<N: NestedValue<D>, D: Domain> AnnotatedValue<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> AnnotatedValue<N, D> {
     fn new(anns: Annotations<N, D>, value: Value<N, D>) -> Self {
         AnnotatedValue(anns, value)
     }
 }
 
-impl<N: NestedValue<D>, D: Domain> PartialEq for AnnotatedValue<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> PartialEq for AnnotatedValue<N, D> {
     fn eq(&self, other: &Self) -> bool {
         self.1.eq(&other.1)
     }
 }
 
-impl<N: NestedValue<D>, D: Domain> Eq for AnnotatedValue<N, D> {}
+impl<N: NestedValue<D>, D: Embeddable> Eq for AnnotatedValue<N, D> {}
 
-impl<N: NestedValue<D>, D: Domain> Hash for AnnotatedValue<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> Hash for AnnotatedValue<N, D> {
     fn hash<H: Hasher>(&self, state: &mut H) {
         self.1.hash(state);
     }
 }
 
-impl<N: NestedValue<D>, D: Domain> PartialOrd for AnnotatedValue<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> PartialOrd for AnnotatedValue<N, D> {
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
         Some(self.cmp(other))
     }
 }
 
-impl<N: NestedValue<D>, D: Domain> Ord for AnnotatedValue<N, D> {
+impl<N: NestedValue<D>, D: Embeddable> Ord for AnnotatedValue<N, D> {
     fn cmp(&self, other: &Self) -> Ordering {
         self.1.cmp(&other.1)
     }
@@ -952,9 +1050,9 @@ impl<N: NestedValue<D>, D: Domain> Ord for AnnotatedValue<N, D> {
 //---------------------------------------------------------------------------
 
 #[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
-pub struct PlainValue<D: Domain>(AnnotatedValue<PlainValue<D>, D>);
+pub struct PlainValue<D: Embeddable>(AnnotatedValue<PlainValue<D>, D>);
 
-impl<D: Domain> PlainValue<D> {
+impl<D: Embeddable> PlainValue<D> {
     pub fn annotations_mut(&mut self) -> &mut Annotations<Self, D> {
         &mut (self.0).0
     }
@@ -964,7 +1062,7 @@ impl<D: Domain> PlainValue<D> {
     }
 }
 
-impl<D: Domain> NestedValue<D> for PlainValue<D> {
+impl<D: Embeddable> NestedValue<D> for PlainValue<D> {
     fn wrap(anns: Annotations<Self, D>, v: Value<Self, D>) -> Self {
         PlainValue(AnnotatedValue::new(anns, v))
     }
@@ -987,32 +1085,20 @@ impl<D: Domain> NestedValue<D> for PlainValue<D> {
     }
 }
 
-impl<D: Domain> Debug for PlainValue<D> {
+impl<D: Embeddable> Debug for PlainValue<D> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         self.debug_fmt(f)
     }
 }
 
-impl<D: Domain> serde::Serialize for PlainValue<D> {
-    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error> where S: serde::Serializer {
-        serialize_nested_value(self, serializer)
-    }
-}
-
-impl<'de, Dom: Domain> serde::Deserialize<'de> for PlainValue<Dom> {
-    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error> where D: serde::Deserializer<'de> {
-        deserialize_nested_value(deserializer)
-    }
-}
-
 //---------------------------------------------------------------------------
 
 use std::rc::Rc;
 
 #[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
-pub struct RcValue<D: Domain>(Rc<AnnotatedValue<RcValue<D>, D>>);
+pub struct RcValue<D: Embeddable>(Rc<AnnotatedValue<RcValue<D>, D>>);
 
-impl<D: Domain> NestedValue<D> for RcValue<D> {
+impl<D: Embeddable> NestedValue<D> for RcValue<D> {
     fn wrap(anns: Annotations<Self, D>, v: Value<Self, D>) -> Self {
         RcValue(Rc::new(AnnotatedValue::new(anns, v)))
     }
@@ -1037,30 +1123,18 @@ impl<D: Domain> NestedValue<D> for RcValue<D> {
     }
 }
 
-impl<D: Domain> Debug for RcValue<D> {
+impl<D: Embeddable> Debug for RcValue<D> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         self.debug_fmt(f)
     }
 }
 
-impl<D: Domain> serde::Serialize for RcValue<D> {
-    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error> where S: serde::Serializer {
-        serialize_nested_value(self, serializer)
-    }
-}
-
-impl<'de, Dom: Domain> serde::Deserialize<'de> for RcValue<Dom> {
-    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error> where D: serde::Deserializer<'de> {
-        deserialize_nested_value(deserializer)
-    }
-}
-
 //---------------------------------------------------------------------------
 
 #[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
-pub struct ArcValue<D: Domain>(Arc<AnnotatedValue<ArcValue<D>, D>>);
+pub struct ArcValue<D: Embeddable>(Arc<AnnotatedValue<ArcValue<D>, D>>);
 
-impl<D: Domain> NestedValue<D> for ArcValue<D> {
+impl<D: Embeddable> NestedValue<D> for ArcValue<D> {
     fn wrap(anns: Annotations<Self, D>, v: Value<Self, D>) -> Self {
         ArcValue(Arc::new(AnnotatedValue::new(anns, v)))
     }
@@ -1085,45 +1159,19 @@ impl<D: Domain> NestedValue<D> for ArcValue<D> {
     }
 }
 
-impl<D: Domain> Debug for ArcValue<D> {
+impl<D: Embeddable> Debug for ArcValue<D> {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         self.debug_fmt(f)
     }
 }
 
-impl<D: Domain> serde::Serialize for ArcValue<D> {
-    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error> where S: serde::Serializer {
-        serialize_nested_value(self, serializer)
-    }
-}
-
-impl<'de, Dom: Domain> serde::Deserialize<'de> for ArcValue<Dom> {
-    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error> where D: serde::Deserializer<'de> {
-        deserialize_nested_value(deserializer)
-    }
-}
-
 //---------------------------------------------------------------------------
 
 #[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
 pub struct IOValue(Arc<AnnotatedValue<IOValue, IOValue>>);
 pub type UnwrappedIOValue = Value<IOValue, IOValue>;
 
-impl Domain for IOValue {
-    fn from_preserves(v: IOValue) -> IOResult<Self> {
-        Ok(v)
-    }
-
-    fn as_preserves(&self) -> IOValue {
-        self.clone()
-    }
-}
-
-lazy_static! {
-    pub static ref FALSE: IOValue = IOValue(Arc::new(AnnotatedValue(Annotations::empty(), Value::Boolean(false))));
-    pub static ref TRUE: IOValue = IOValue(Arc::new(AnnotatedValue(Annotations::empty(), Value::Boolean(true))));
-    pub static ref EMPTY_SEQ: IOValue = IOValue(Arc::new(AnnotatedValue(Annotations::empty(), Value::Sequence(Vec::new()))));
-}
+impl Domain for IOValue {}
 
 impl NestedValue<IOValue> for IOValue {
     fn wrap(anns: Annotations<Self, IOValue>, v: Value<Self, IOValue>) -> Self {
@@ -1151,14 +1199,6 @@ impl NestedValue<IOValue> for IOValue {
             Err(r) => r.1.clone(),
         }
     }
-
-    fn to_io_value(&self) -> IOValue {
-        self.clone()
-    }
-
-    fn from_io_value(v: &IOValue) -> IOResult<Self> {
-        Ok(v.clone())
-    }
 }
 
 impl Debug for IOValue {
@@ -1169,12 +1209,52 @@ impl Debug for IOValue {
 
 impl serde::Serialize for IOValue {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error> where S: serde::Serializer {
-        serialize_nested_value(self, serializer)
+        super::magic::output_value(serializer, self.clone())
     }
 }
 
 impl<'de> serde::Deserialize<'de> for IOValue {
     fn deserialize<D>(deserializer: D) -> Result<Self, D::Error> where D: serde::Deserializer<'de> {
-        deserialize_nested_value(deserializer)
+        super::magic::input_value(deserializer)
+    }
+}
+
+//---------------------------------------------------------------------------
+
+#[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
+pub struct DummyValue<D: Embeddable>(AnnotatedValue<DummyValue<D>, D>);
+
+impl<D: Embeddable> Debug for DummyValue<D> {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        f.write_str("<>")
+    }
+}
+
+
+impl<D: Embeddable> DummyValue<D> {
+    pub fn new() -> Self {
+        DummyValue(AnnotatedValue::new(Annotations::empty(), Value::Boolean(false)))
+    }
+}
+
+impl<D: Embeddable> NestedValue<D> for DummyValue<D> {
+    fn wrap(_anns: Annotations<Self, D>, _v: Value<Self, D>) -> Self {
+        DummyValue::new()
+    }
+
+    fn annotations(&self) -> &Annotations<Self, D> {
+        &self.0.0
+    }
+
+    fn value(&self) -> &Value<Self, D> {
+        &self.0.1
+    }
+
+    fn pieces(self) -> (Annotations<Self, D>, Value<Self, D>) {
+        (self.0.0, self.0.1)
+    }
+
+    fn value_owned(self) -> Value<Self, D> {
+        self.0.1
     }
 }
